@@ -22,6 +22,8 @@
 ContextBroker routines for Python:
   - cbManager.send_batch
   - cbManager.get_entities_page
+  - cbManager.get_entities
+  - cbManager.delete_entities
 '''
 import requests
 from requests.adapters import HTTPAdapter
@@ -34,6 +36,9 @@ import json
 
 from . import authManager
 
+# control urllib3 post and get verify in false
+import urllib3, urllib3.exceptions
+urllib3.disable_warnings(category=urllib3.exceptions.InsecureRequestWarning)
 
 logger = logging.getLogger(__name__)
 
@@ -62,6 +67,8 @@ class cbManager:
     post_retry_backoff_factor: retry factor delay -> {backoff factor} * (2 ** ({number of total retries} - 1)) (defaul: 20). (ref. urllib3 docs) (default: False)
     sleep_send_batch: sleep X seconds afters send update batch. (default: 0). 
     cb_flowcontrol: Opción del Context Broker, que permite un mejor rendimiento en caso de envío masivo de datos (batch updates). Este mecanismo, requiere arrancar el Context Broker con un flag concreto y en las peticiones de envío de datos, añadir esa opción. Referencia en Fiware Orion Docs (default: False)
+    block_size: maximum size per batch, in bytes. Default is 800kb and it is not recommended to change.
+    batch_size: maximum size per batch, in entites. Default is 0 (no limitiation, other than block_size).
     """
     
     endpoint: str
@@ -75,8 +82,14 @@ class cbManager:
     # see https://fiware-orion.readthedocs.io/en/master/user/known_limitations.html)
     # and it is not recommended to change it
     block_size = 800000
+    # Batch size is entities. Ignored when batch_size = 0.
+    # If > 0 , limits the number of entities per batch.
+    # We might want to limit the amount of entities per batch for
+    # reasons other than block size (e.g to avoid triggering too
+    # many subscriptions per batch and stressing cygnus queues)
+    batch_size = 0
     
-    def __init__(self,*, endpoint: str = None, timeout: int = 10, post_retry_connect: int = 3, post_retry_backoff_factor: int = 20, sleep_send_batch: int = 0, cb_flowcontrol: bool = False, block_size: int = 800000) -> None:
+    def __init__(self,*, endpoint: str = None, timeout: int = 10, post_retry_connect: int = 3, post_retry_backoff_factor: int = 20, sleep_send_batch: int = 0, cb_flowcontrol: bool = False, block_size: int = 800000, batch_size: int = 0) -> None:
         
         if (endpoint == None):
             raise ValueError(f'You must define <<endpoint>> in cbManager')
@@ -87,6 +100,7 @@ class cbManager:
         self.post_retry_backoff_factor = post_retry_backoff_factor
         self.sleep_send_batch = sleep_send_batch
         self.cb_flowcontrol = cb_flowcontrol
+        self.batch_size = batch_size
         
         # check block_size limit.
         if (int(block_size) > int(800000)):
@@ -95,7 +109,7 @@ class cbManager:
         self.block_size = block_size
 
 
-    def delete_entities(self, *, service: str = None, subservice: str = None, auth: authManager = None, limit: int = 100, type: str = None, q: str = None, mq: str = None, georel: str = None, geometry: str = None, coords: str = None, id: str = None):
+    def delete_entities(self, *, service: str = None, subservice: str = None, auth: authManager = None, limit: int = 100, type: str = None, q: str = None, mq: str = None, georel: str = None, geometry: str = None, coords: str = None, id: str = None, options_get: list = [], options_send: list = []):
         """Delete data from context broker
 
         :param service: Define service from which entities are deleted, defaults to None
@@ -109,10 +123,12 @@ class cbManager:
         :param geometry: Allows to define the reference shape to be used when resolving the query. (point | polygon | line | box), defaults to None
         :param coords: Must be a string containing a semicolon-separated list of pairs of geographical coordinates in accordance with the geometry specified, defaults to None
         :param id: Delete entities filtering by Identity, defaults to None
+        :param options_get: Options used in Context Broker to find entities, defaults to None
+        :param options_send: Options used in Context Broker to delete entities, defaults to None
         :raises ValueError: is thrown when some required argument is missing
         :raises FetchError: is thrown when the response from the cb indicates an error
         """
-        data = self.get_entities(service=service, subservice=subservice, auth=auth, limit = limit, type = type, q = q, mq = mq, georel = georel, geometry = geometry, coords = coords, id = id)
+        data = self.get_entities(service=service, subservice=subservice, auth=auth, limit = limit, type = type, q = q, mq = mq, georel = georel, geometry = geometry, coords = coords, id = id, options=options_get)
 
         entities = []
         for i, item in enumerate(data):
@@ -122,9 +138,9 @@ class cbManager:
             }
             entities.append(entity)
         
-        self.send_batch(service=service, subservice=subservice, auth=auth, entities=entities, actionType='delete')
+        self.send_batch(service=service, subservice=subservice, auth=auth, entities=entities, actionType='delete', options=options_send)
 
-    def get_entities(self, *, service: str = None, subservice: str = None, auth: authManager = None, limit: int = 100, type: str = None, orderBy: str = None, q: str = None, mq: str = None, georel: str = None, geometry: str = None, coords: str = None, id: str = None):
+    def get_entities(self, *, service: str = None, subservice: str = None, auth: authManager = None, limit: int = 100, type: str = None, orderBy: str = None, q: str = None, mq: str = None, georel: str = None, geometry: str = None, coords: str = None, id: str = None, options: list = []):
         """Retrieve data from context broker
 
         :param service: Define service from which entities are retrieved, defaults to None
@@ -139,6 +155,7 @@ class cbManager:
         :param geometry: Allows to define the reference shape to be used when resolving the query. (point | polygon | line | box), defaults to None
         :param coords: Must be a string containing a semicolon-separated list of pairs of geographical coordinates in accordance with the geometry specified, defaults to None
         :param id: Retrieve entities filtering by Identity, defaults to None
+        :param options: Options used to retrive entities, defaults to None
         :raises ValueError: is thrown when some required argument is missing
         :raises FetchError: is thrown when the response from the cb indicates an error
         :return: json data
@@ -150,12 +167,12 @@ class cbManager:
         data = ['go!']
         while (data != []) :
             offset = (pg-1)*limit
-            data = self.get_entities_page(service=service, subservice=subservice, auth=auth, offset = offset, limit = limit, type = type, orderBy = orderBy, q = q, mq = mq, georel = georel, geometry = geometry, coords = coords, id = id)
+            data = self.get_entities_page(service=service, subservice=subservice, auth=auth, offset = offset, limit = limit, type = type, orderBy = orderBy, q = q, mq = mq, georel = georel, geometry = geometry, coords = coords, id = id, options = options)
             pg += 1
             result += data
         return result
 
-    def get_entities_page(self, *, service:str = None, subservice: str = None, auth: authManager = None, offset: int = None, limit: int = None, type: str = None, orderBy: str = None, q: str = None, mq: str = None, georel: str = None, geometry: str = None, coords: str = None, id: str = None):
+    def get_entities_page(self, *, service:str = None, subservice: str = None, auth: authManager = None, offset: int = None, limit: int = None, type: str = None, orderBy: str = None, q: str = None, mq: str = None, georel: str = None, geometry: str = None, coords: str = None, id: str = None, options: list = []):
         """Retrieve data from context broker
 
         :param service: Define service from which entities are retrieved, defaults to None or auth.service defined value
@@ -171,6 +188,7 @@ class cbManager:
         :param geometry: Allows to define the reference shape to be used when resolving the query. (point | polygon | line | box), defaults to None
         :param coords: Must be a string containing a semicolon-separated list of pairs of geographical coordinates in accordance with the geometry specified, defaults to None
         :param id: Retrieve entities filtering by Identity, defaults to None
+        :param options: Options used, defaults to None
         :raises ValueError: is thrown when some required argument is missing
         :raises FetchError: is thrown when the response from the cb indicates an error
         :return: json data
@@ -216,8 +234,15 @@ class cbManager:
             else:
                 raise ValueError('If use geographical queries, you must define georel, geometry and coords in params')
         
+        
         params = {"offset": offset, "limit": limit, "type": type, "orderBy": orderBy, "q": q, "mq": mq, "georel": georel, "geometry": geometry, "coords": coords, "id": id}
-        req_url = f"{self.endpoint}/v2/entities"
+        
+        req_url = ""
+        if (options != None and len(options) > 0):
+            req_url = f"{self.endpoint}/v2/entities?options={','.join(options)}"
+        else:    
+            req_url = f"{self.endpoint}/v2/entities"
+        
         resp = requests.get(req_url, params=params, headers=headers, verify=False, timeout=self.timeout)
         if resp.status_code == 400 or resp.status_code == 401:
             respjson = resp.json()
@@ -228,7 +253,7 @@ class cbManager:
         return resp.json()
 
 
-    def send_batch(self, *, service:str = None, subservice: str = None, auth: authManager = None, entities: str, actionType: str = 'append') -> bool:
+    def send_batch(self, *, service:str = None, subservice: str = None, auth: authManager = None, entities: str, actionType: str = 'append', options: list = []) -> bool:
         """Send batch data to context broker with block control
 
         :param auth: Define authManager 
@@ -236,6 +261,7 @@ class cbManager:
         :param service: Define service to send batch data, defaults to None
         :param subservice: Define subservice to send batch data, defaults to None
         :param actionType: Batch action type, defaults is append
+        :param options: Options used, defaults to None
         :raises ValueError: is thrown when some required argument is missing
         :raises Exception: is thrown when the cotext broker response isn't ok operation
         :return: True if the operation is correct
@@ -246,18 +272,18 @@ class cbManager:
             entitiesToSend.append(entity)
             accumulated_block += len(json.dumps(entity))
             
-            if accumulated_block > self.block_size:
+            if accumulated_block > self.block_size or (self.batch_size > 0 and len(entitiesToSend) >= self.batch_size) :
                 logger.debug(f'- Sending a batch {actionType} of {len(entitiesToSend)} entities')
-                self.__send_batch(auth=auth, service=service, subservice=subservice, entities=entitiesToSend, actionType=actionType)
+                self.__send_batch(auth=auth, service=service, subservice=subservice, entities=entitiesToSend, actionType=actionType, options=options)
                 entitiesToSend = []
                 accumulated_block = 0
         
         # Remaining block, if any
         if accumulated_block > 0:
             logger.debug(f'- Sending final batch {actionType} of {len(entitiesToSend)} entities')
-            self.__send_batch(auth=auth, service=service, subservice=subservice, entities=entitiesToSend, actionType=actionType)
+            self.__send_batch(auth=auth, service=service, subservice=subservice, entities=entitiesToSend, actionType=actionType, options=options)
 
-    def __send_batch(self, *, service:str = None, subservice: str = None, auth: authManager = None, entities: str, actionType: str = 'append') -> bool:
+    def __send_batch(self, *, service:str = None, subservice: str = None, auth: authManager = None, entities: str, actionType: str = 'append', options: list = []) -> bool:
         """Send batch data to context broker
 
         :param auth: Define authManager 
@@ -265,6 +291,7 @@ class cbManager:
         :param service: Define service to send batch data, defaults to None
         :param subservice: Define subservice to send batch data, defaults to None  or auth.subservice defined value
         :param actionType: Batch action type, defaults is append
+        :param options: Options used, defaults to None
         :raises ValueError: is thrown when some required argument is missing
         :raises Exception: is thrown when the cotext broker response isn't ok operation
         :return: True if the operation is correct
@@ -286,10 +313,10 @@ class cbManager:
         if (auth != None and subservice not in auth.tokens.keys()):
             auth.get_auth_token_subservice(subservice = subservice)
 
-        res = self.__batch_creation(auth=auth, service=service, subservice = subservice, entities=entities, actionType=actionType)
+        res = self.__batch_creation(auth=auth, service=service, subservice = subservice, entities=entities, actionType=actionType, options=options)
         if (auth != None and res.status_code == 401):
             auth.get_auth_token_subservice(subservice = subservice)
-            res = self.__batch_creation(auth=auth, service=service, subservice = subservice, entities=entities, actionType=actionType)
+            res = self.__batch_creation(auth=auth, service=service, subservice = subservice, entities=entities, actionType=actionType, options=options)
         
         if res.status_code != 204:
             raise Exception(f'Error in batch {actionType} operation ({res.status_code}): {res.json()}')
@@ -301,7 +328,7 @@ class cbManager:
             
         return True 
         
-    def __batch_creation(self, *, service: str = None, subservice: str = None, auth: authManager = None, entities: str, actionType: str = 'append'):
+    def __batch_creation(self, *, service: str = None, subservice: str = None, auth: authManager = None, entities: str, actionType: str = 'append', options: list = []):
         """Send batch data to Context Broker
 
         :param entities: Entities data
@@ -309,6 +336,7 @@ class cbManager:
         :param subservice: Define subservice to send batch data, defaults to None or auth.subservice defined value
         :param auth: Define authManager, defaults to None
         :param actionType: Batch action type, defaults is append
+        :params options: Options used, default to None
         :raises ValueError: is thrown when some required argument is missing
         :return: Http response code
         """
@@ -349,11 +377,20 @@ class cbManager:
             'actionType': f'{actionType}',
             'entities': entities
         }
-
+        
+        # if cb_flowcontrol, add flowControl flag to options
         if (self.cb_flowcontrol):
-            req_url = f'{self.endpoint}/v2/op/update?options=flowControl'
-        else:
-            req_url = f'{self.endpoint}/v2/op/update'
+            if (options == None):
+                options = ['flowControl']
+            else:
+                # check if flowcontrol is in options.
+                if 'flowControl' not in options:
+                    options.append('flowControl')
+
+        if (options != None and len(options) > 0):
+            req_url = f"{self.endpoint}/v2/op/update?options={','.join(options)}"
+        else:    
+            req_url = f"{self.endpoint}/v2/op/update"
     
         http = requests.Session()
         retry_strategy = Retry(
